@@ -1,5 +1,8 @@
 //! Error types
-use crate::primitive_type::{PrimitiveType, PrimitiveTypesBitMap};
+use crate::{
+    paths::JSONPointer,
+    primitive_type::{PrimitiveType, PrimitiveTypesBitMap},
+};
 use serde_json::{Map, Number, Value};
 use std::{
     borrow::Cow,
@@ -51,7 +54,7 @@ pub struct ValidationError<'a> {
     /// Type of validation error
     pub kind: ValidationErrorKind,
     /// Path of the property that failed validation
-    pub instance_path: Vec<String>,
+    pub instance_path: JSONPointer,
 }
 
 /// An iterator over instances of `ValidationError` that represent validation error for the
@@ -91,6 +94,8 @@ pub(crate) fn error(instance: ValidationError) -> ErrorIterator {
 pub enum ValidationErrorKind {
     /// The input array contain more items than expected.
     AdditionalItems { limit: usize },
+    /// Unexpected properties.
+    AdditionalProperties { unexpected: Vec<String> },
     /// The input value is not valid under any of the given schemas.
     AnyOf,
     /// The input value doesn't match expected constant.
@@ -149,6 +154,10 @@ pub enum ValidationErrorKind {
     OneOfNotValid,
     /// When the input doesn't match to a pattern.
     Pattern { pattern: String },
+    /// Object property names are invalid.
+    PropertyNames {
+        error: Box<ValidationError<'static>>,
+    },
     /// When a required property is missing.
     Required { property: String },
     /// Any error that happens during network request via `reqwest` crate
@@ -182,7 +191,7 @@ impl<'a> ValidationError<'a> {
     }
 
     pub(crate) fn additional_items(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         limit: usize,
     ) -> ValidationError<'a> {
@@ -192,7 +201,18 @@ impl<'a> ValidationError<'a> {
             kind: ValidationErrorKind::AdditionalItems { limit },
         }
     }
-    pub(crate) fn any_of(instance_path: Vec<String>, instance: &'a Value) -> ValidationError<'a> {
+    pub(crate) fn additional_properties(
+        instance_path: JSONPointer,
+        instance: &'a Value,
+        unexpected: Vec<String>,
+    ) -> ValidationError<'a> {
+        ValidationError {
+            instance_path,
+            instance: Cow::Borrowed(instance),
+            kind: ValidationErrorKind::AdditionalProperties { unexpected },
+        }
+    }
+    pub(crate) fn any_of(instance_path: JSONPointer, instance: &'a Value) -> ValidationError<'a> {
         ValidationError {
             instance_path,
             instance: Cow::Borrowed(instance),
@@ -200,7 +220,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn constant_array(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         expected_value: &[Value],
     ) -> ValidationError<'a> {
@@ -213,7 +233,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn constant_boolean(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         expected_value: bool,
     ) -> ValidationError<'a> {
@@ -226,7 +246,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn constant_null(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
     ) -> ValidationError<'a> {
         ValidationError {
@@ -238,7 +258,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn constant_number(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         expected_value: &Number,
     ) -> ValidationError<'a> {
@@ -251,7 +271,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn constant_object(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         expected_value: &Map<String, Value>,
     ) -> ValidationError<'a> {
@@ -264,7 +284,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn constant_string(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         expected_value: &str,
     ) -> ValidationError<'a> {
@@ -276,7 +296,7 @@ impl<'a> ValidationError<'a> {
             },
         }
     }
-    pub(crate) fn contains(instance_path: Vec<String>, instance: &'a Value) -> ValidationError<'a> {
+    pub(crate) fn contains(instance_path: JSONPointer, instance: &'a Value) -> ValidationError<'a> {
         ValidationError {
             instance_path,
             instance: Cow::Borrowed(instance),
@@ -284,7 +304,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn content_encoding(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         encoding: &str,
     ) -> ValidationError<'a> {
@@ -297,7 +317,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn content_media_type(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         media_type: &str,
     ) -> ValidationError<'a> {
@@ -310,7 +330,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn enumeration(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         options: &Value,
     ) -> ValidationError<'a> {
@@ -323,7 +343,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn exclusive_maximum(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         limit: f64,
     ) -> ValidationError<'a> {
@@ -334,7 +354,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn exclusive_minimum(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         limit: f64,
     ) -> ValidationError<'a> {
@@ -345,7 +365,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn false_schema(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
     ) -> ValidationError<'a> {
         ValidationError {
@@ -356,13 +376,13 @@ impl<'a> ValidationError<'a> {
     }
     pub(crate) fn file_not_found(error: io::Error) -> ValidationError<'a> {
         ValidationError {
-            instance_path: Vec::new(),
+            instance_path: JSONPointer::default(),
             instance: Cow::Owned(Value::Null),
             kind: ValidationErrorKind::FileNotFound { error },
         }
     }
     pub(crate) fn format(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         format: &'static str,
     ) -> ValidationError<'a> {
@@ -374,34 +394,34 @@ impl<'a> ValidationError<'a> {
     }
     pub(crate) fn from_utf8(error: FromUtf8Error) -> ValidationError<'a> {
         ValidationError {
-            instance_path: Vec::new(),
+            instance_path: JSONPointer::default(),
             instance: Cow::Owned(Value::Null),
             kind: ValidationErrorKind::FromUtf8 { error },
         }
     }
     pub(crate) fn json_parse(error: serde_json::Error) -> ValidationError<'a> {
         ValidationError {
-            instance_path: Vec::new(),
+            instance_path: JSONPointer::default(),
             instance: Cow::Owned(Value::Null),
             kind: ValidationErrorKind::JSONParse { error },
         }
     }
     pub(crate) fn invalid_reference(reference: String) -> ValidationError<'a> {
         ValidationError {
-            instance_path: Vec::new(),
+            instance_path: JSONPointer::default(),
             instance: Cow::Owned(Value::Null),
             kind: ValidationErrorKind::InvalidReference { reference },
         }
     }
     pub(crate) fn invalid_url(error: url::ParseError) -> ValidationError<'a> {
         ValidationError {
-            instance_path: Vec::new(),
+            instance_path: JSONPointer::default(),
             instance: Cow::Owned(Value::Null),
             kind: ValidationErrorKind::InvalidURL { error },
         }
     }
     pub(crate) fn max_items(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         limit: u64,
     ) -> ValidationError<'a> {
@@ -412,7 +432,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn maximum(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         limit: f64,
     ) -> ValidationError<'a> {
@@ -423,7 +443,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn max_length(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         limit: u64,
     ) -> ValidationError<'a> {
@@ -434,7 +454,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn max_properties(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         limit: u64,
     ) -> ValidationError<'a> {
@@ -445,7 +465,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn min_items(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         limit: u64,
     ) -> ValidationError<'a> {
@@ -456,7 +476,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn minimum(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         limit: f64,
     ) -> ValidationError<'a> {
@@ -467,7 +487,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn min_length(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         limit: u64,
     ) -> ValidationError<'a> {
@@ -478,7 +498,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn min_properties(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         limit: u64,
     ) -> ValidationError<'a> {
@@ -489,7 +509,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn multiple_of(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         multiple_of: f64,
     ) -> ValidationError<'a> {
@@ -500,7 +520,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn not(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         schema: Value,
     ) -> ValidationError<'a> {
@@ -511,7 +531,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn one_of_multiple_valid(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
     ) -> ValidationError<'a> {
         ValidationError {
@@ -521,7 +541,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn one_of_not_valid(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
     ) -> ValidationError<'a> {
         ValidationError {
@@ -531,7 +551,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn pattern(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         pattern: String,
     ) -> ValidationError<'a> {
@@ -541,8 +561,21 @@ impl<'a> ValidationError<'a> {
             kind: ValidationErrorKind::Pattern { pattern },
         }
     }
+    pub(crate) fn property_names(
+        instance_path: JSONPointer,
+        instance: &'a Value,
+        error: ValidationError<'a>,
+    ) -> ValidationError<'a> {
+        ValidationError {
+            instance_path,
+            instance: Cow::Borrowed(instance),
+            kind: ValidationErrorKind::PropertyNames {
+                error: Box::new(error.into_owned()),
+            },
+        }
+    }
     pub(crate) fn required(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         property: String,
     ) -> ValidationError<'a> {
@@ -555,20 +588,20 @@ impl<'a> ValidationError<'a> {
     #[cfg(any(feature = "reqwest", test))]
     pub(crate) fn reqwest(error: reqwest::Error) -> ValidationError<'a> {
         ValidationError {
-            instance_path: Vec::new(),
+            instance_path: JSONPointer::default(),
             instance: Cow::Owned(Value::Null),
             kind: ValidationErrorKind::Reqwest { error },
         }
     }
     pub(crate) fn schema() -> ValidationError<'a> {
         ValidationError {
-            instance_path: Vec::new(),
+            instance_path: JSONPointer::default(),
             instance: Cow::Owned(Value::Null),
             kind: ValidationErrorKind::Schema,
         }
     }
     pub(crate) fn single_type_error(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         type_name: PrimitiveType,
     ) -> ValidationError<'a> {
@@ -581,7 +614,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn multiple_type_error(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
         types: PrimitiveTypesBitMap,
     ) -> ValidationError<'a> {
@@ -594,7 +627,7 @@ impl<'a> ValidationError<'a> {
         }
     }
     pub(crate) fn unique_items(
-        instance_path: Vec<String>,
+        instance_path: JSONPointer,
         instance: &'a Value,
     ) -> ValidationError<'a> {
         ValidationError {
@@ -605,14 +638,14 @@ impl<'a> ValidationError<'a> {
     }
     pub(crate) fn unknown_reference_scheme(scheme: String) -> ValidationError<'a> {
         ValidationError {
-            instance_path: Vec::new(),
+            instance_path: JSONPointer::default(),
             instance: Cow::Owned(Value::Null),
             kind: ValidationErrorKind::UnknownReferenceScheme { scheme },
         }
     }
     pub(crate) fn utf8(error: Utf8Error) -> ValidationError<'a> {
         ValidationError {
-            instance_path: Vec::new(),
+            instance_path: JSONPointer::default(),
             instance: Cow::Owned(Value::Null),
             kind: ValidationErrorKind::Utf8 { error },
         }
@@ -710,6 +743,25 @@ impl fmt::Display for ValidationError<'_> {
                     verb
                 )
             }
+            ValidationErrorKind::AdditionalProperties { unexpected } => {
+                let verb = {
+                    if unexpected.len() == 1 {
+                        "was"
+                    } else {
+                        "were"
+                    }
+                };
+                write!(
+                    f,
+                    "Additional properties are not allowed ({} {} unexpected)",
+                    unexpected
+                        .iter()
+                        .map(|x| format!("'{}'", x))
+                        .collect::<Vec<String>>()
+                        .join(", "),
+                    verb
+                )
+            }
             ValidationErrorKind::AnyOf | ValidationErrorKind::OneOfNotValid => write!(
                 f,
                 "'{}' is not valid under any of the given schemas",
@@ -819,6 +871,9 @@ impl fmt::Display for ValidationError<'_> {
             ValidationErrorKind::Pattern { pattern } => {
                 write!(f, "'{}' does not match '{}'", self.instance, pattern)
             }
+            ValidationErrorKind::PropertyNames { error } => {
+                write!(f, "{}", error.to_string())
+            }
             ValidationErrorKind::Required { property } => {
                 write!(f, "'{}' is a required property", property)
             }
@@ -850,14 +905,18 @@ impl fmt::Display for ValidationError<'_> {
 #[cfg(test)]
 mod tests {
     use super::*;
-    use crate::JSONSchema;
+    use crate::{paths::PathChunk, JSONSchema};
     use serde_json::json;
     use test_case::test_case;
 
     #[test]
     fn single_type_error() {
         let instance = json!(42);
-        let err = ValidationError::single_type_error(vec![], &instance, PrimitiveType::String);
+        let err = ValidationError::single_type_error(
+            JSONPointer::default(),
+            &instance,
+            PrimitiveType::String,
+        );
         assert_eq!(err.to_string(), "'42' is not of type 'string'")
     }
 
@@ -865,7 +924,7 @@ mod tests {
     fn multiple_types_error() {
         let instance = json!(42);
         let err = ValidationError::multiple_type_error(
-            vec![],
+            JSONPointer::default(),
             &instance,
             vec![PrimitiveType::String, PrimitiveType::Number].into(),
         );
@@ -898,17 +957,17 @@ mod tests {
         let error = result.next().expect("validation error");
 
         assert!(result.next().is_none());
-        assert_eq!(error.instance_path, expected);
+        assert_eq!(error.instance_path, JSONPointer::from(expected));
     }
 
-    #[test_case(true, &json!([1, {"foo": ["42"]}]), &["0"])]
-    #[test_case(true, &json!(["a", {"foo": [42]}]), &["1", "foo", "0"])]
-    #[test_case(false, &json!([1, {"foo": ["42"]}]), &["0"])]
-    #[test_case(false, &json!(["a", {"foo": [42]}]), &["1", "foo", "0"])]
+    #[test_case(true, &json!([1, {"foo": ["42"]}]), &[PathChunk::Index(0)])]
+    #[test_case(true, &json!(["a", {"foo": [42]}]), &[PathChunk::Index(1), PathChunk::Name("foo".to_string()), PathChunk::Index(0)])]
+    #[test_case(false, &json!([1, {"foo": ["42"]}]), &[PathChunk::Index(0)])]
+    #[test_case(false, &json!(["a", {"foo": [42]}]), &[PathChunk::Index(1), PathChunk::Name("foo".to_string()), PathChunk::Index(0)])]
     fn instance_path_properties_and_arrays(
         additional_items: bool,
         instance: &Value,
-        expected: &[&str],
+        expected: &[PathChunk],
     ) {
         let schema = json!(
             {
@@ -939,14 +998,18 @@ mod tests {
         let error = result.next().expect("validation error");
 
         assert!(result.next().is_none());
-        assert_eq!(error.instance_path, expected);
+        assert_eq!(error.instance_path, JSONPointer::from(expected));
     }
 
-    #[test_case(true, &json!([[1, 2, 3], [4, "5", 6], [7, 8, 9]]), &["1", "1"])]
-    #[test_case(false, &json!([[1, 2, 3], [4, "5", 6], [7, 8, 9]]), &["1", "1"])]
-    #[test_case(true, &json!([[1, 2, 3], [4, 5, 6], 42]), &["2"])]
-    #[test_case(false, &json!([[1, 2, 3], [4, 5, 6], 42]), &["2"])]
-    fn instance_path_nested_arrays(additional_items: bool, instance: &Value, expected: &[&str]) {
+    #[test_case(true, &json!([[1, 2, 3], [4, "5", 6], [7, 8, 9]]), &[PathChunk::Index(1), PathChunk::Index(1)])]
+    #[test_case(false, &json!([[1, 2, 3], [4, "5", 6], [7, 8, 9]]), &[PathChunk::Index(1), PathChunk::Index(1)])]
+    #[test_case(true, &json!([[1, 2, 3], [4, 5, 6], 42]), &[PathChunk::Index(2)])]
+    #[test_case(false, &json!([[1, 2, 3], [4, 5, 6], 42]), &[PathChunk::Index(2)])]
+    fn instance_path_nested_arrays(
+        additional_items: bool,
+        instance: &Value,
+        expected: &[PathChunk],
+    ) {
         let schema = json!(
             {
                 "additionalItems": additional_items,
@@ -964,14 +1027,14 @@ mod tests {
         let error = result.next().expect("validation error");
 
         assert!(result.next().is_none());
-        assert_eq!(error.instance_path, expected);
+        assert_eq!(error.instance_path, JSONPointer::from(expected));
     }
 
-    #[test_case(true, &json!([1, "a"]), &["1"])]
-    #[test_case(false, &json!([1, "a"]), &["1"])]
+    #[test_case(true, &json!([1, "a"]), &[PathChunk::Index(1)])]
+    #[test_case(false, &json!([1, "a"]), &[PathChunk::Index(1)])]
     #[test_case(true, &json!(123), &[])]
     #[test_case(false, &json!(123), &[])]
-    fn instance_path_arrays(additional_items: bool, instance: &Value, expected: &[&str]) {
+    fn instance_path_arrays(additional_items: bool, instance: &Value, expected: &[PathChunk]) {
         let schema = json!(
             {
                 "additionalItems": additional_items,
@@ -986,6 +1049,6 @@ mod tests {
         let error = result.next().expect("validation error");
 
         assert!(result.next().is_none());
-        assert_eq!(error.instance_path, expected);
+        assert_eq!(error.instance_path, JSONPointer::from(expected));
     }
 }
